from diot import Diot
from .cmdy_defaults import STDOUT, STDERR


class CmdyActionError(Exception):
    """Wrong actions taken"""


class CmdyTimeoutError(Exception):
    """Timeout running command"""


class CmdyExecNotFoundError(Exception):
    """Unable to find the executable"""


class CmdyReturnCodeError(Exception):
    """Unexpected return code"""

    @staticmethod
    def _out_nowait(result, which):
        if (
            which == STDOUT
            and getattr(result, "_stdout_str", None) is not None
        ):
            return result._stdout_str.splitlines()
        if (
            which == STDERR
            and getattr(result, "_stderr_str", None) is not None
        ):
            return result._stderr_str.splitlines()

        out = result.stdout if which == STDOUT else result.stderr
        if isinstance(out, (str, bytes)):
            return out.splitlines()

        return list(out)

    def __init__(self, result):
        # We cann't do isinstance check for CmdyResult, since it can be from
        # a baked module
        if (
            isinstance(result, Diot)
            or result.__class__.__name__ == "CmdyResult"
        ):

            msgs = [
                f"Unexpected RETURN CODE {result.rc}, "
                f"expecting: {result.holding.okcode}",
                "",
                f"  [   PID] {result.pid}",
                "",
                "  [   CMD] "
                f'{getattr(result, "piped_strcmds", result.cmd)}',
                "",
            ]

            if result.stdout is None:
                msgs.append("  [STDOUT] <NA / ITERATED / REDIRECTED>")
                msgs.append("")
            else:
                outs = CmdyReturnCodeError._out_nowait(result, STDOUT) or [""]
                msgs.append(f"  [STDOUT] {outs.pop().rstrip()}")
                msgs.extend(f"           {out}" for out in outs[:31])
                if len(outs) > 31:
                    msgs.append(f"           [{len(outs)-31} lines hidden.]")
                msgs.append("")

            if result.stderr is None:
                msgs.append("  [STDERR] <NA / ITERATED / REDIRECTED>")
                msgs.append("")
            else:
                errs = CmdyReturnCodeError._out_nowait(result, STDERR) or [""]
                msgs.append(f"  [STDERR] {errs.pop().rstrip()}")
                msgs.extend(f"           {err}" for err in errs[:31])
                if len(errs) > 31:
                    msgs.append(f"           [{len(errs)-31} lines hidden.]")
                msgs.append("")
        else:  # pragma: no cover
            msgs = [str(result)]
        super().__init__("\n".join(msgs))



import sys
import fileinput
import warnings
from typing import TYPE_CHECKING, Union

import curio
from curio import subprocess

from ..cmdy_exceptions import CmdyTimeoutError

if TYPE_CHECKING:
    from ..cmdy_bakeable import Bakeable


def vendor(bakeable: "Bakeable"):
    """Vendor the plugins with the bakeable._plugin_factory"""

    @bakeable._plugin_factory.register
    class PluginFg:
        """Plugin: fg
        Running command in foreground
        Using sys.stdout and sys.stderr"""

        async def _feed(
            self: bakeable.CmdyResult,  # type: ignore
            poll_interval: float,
        ):
            """Try to feed stdout/stderr to sys.stdout/sys.stderr"""

            async def _feed_one(instream, outstream):
                try:
                    out = await curio.timeout_after(
                        poll_interval, instream.__anext__
                    )
                except curio.TaskTimeout:
                    pass
                except StopAsyncIteration:
                    return False
                else:
                    if self.holding.encoding:
                        outstream.write(out.decode(self.holding.encoding))
                    else:
                        outstream.buffer.write(out)
                    outstream.flush()
                return True

            out_live = err_live = True
            while out_live or err_live:
                if out_live:
                    out_live = await _feed_one(self.proc.stdout, sys.stdout)
                if err_live:
                    err_live = await _feed_one(self.proc.stderr, sys.stderr)

            if isinstance(self, bakeable.CmdyAsyncResult):
                await self.wait()

        async def _timeout_wrapper(
            self: bakeable.CmdyResult,  # type: ignore
            poll_interval: float,
        ):
            if self.holding.timeout:
                try:
                    await curio.timeout_after(
                        self.holding.timeout,
                        PluginFg._feed,
                        self,
                        poll_interval,
                    )
                except curio.TaskTimeout:
                    raise CmdyTimeoutError(
                        f"Timeout after {self.holding.timeout} seconds."
                    ) from None
            else:
                await PluginFg._feed(self, poll_interval)

        @bakeable._plugin_factory.hold_then("fg", final=True, hold_right=False)
        def foreground(
            self, stdin: bool = False, poll_interval: Union[bool, float] = 0.1
        ):
            """Running command in foreground
            Using sys.stdout and sys.stderr"""
            self.data.foreground.stdin = stdin
            self.data.foreground.poll_interval = poll_interval

            if not self._onhold():
                return self.run()
            return self

        @bakeable._plugin_factory.add_method(bakeable.CmdyHolding)
        def run(self, wait=None):
            """Run the command and bump stdout/stderr to sys'"""
            orig_run = self._original("run")

            if not self.data.get("foreground"):
                return orig_run(self, wait)

            if (
                (self.data.foreground.stdin and self.stdin != subprocess.PIPE)
                or self.stdout != subprocess.PIPE
                or self.stderr != subprocess.PIPE
            ):
                warnings.warn("Previous redirected pipe will be ignored.")

            # fileinput.input is good to use here
            # as it has fileno()
            self.stdin = (
                fileinput.input() if self.data.foreground.stdin else self.stdin
            )
            self.stdout = subprocess.PIPE
            self.stderr = subprocess.PIPE

            ret = orig_run(self, False)

            # we should handle timeout here, since we are not waiting
            curio.run(
                PluginFg._timeout_wrapper(
                    ret, self.data.foreground.poll_interval
                )
            )
            # we can't in self.wait() in curio.run, because there is
            # already a curio kernel running inside CmdyResult.wait()
            return (
                ret
                if isinstance(ret, bakeable.CmdyAsyncResult)
                else ret.wait()
            )

    return PluginFg()



from importlib import import_module
from pathlib import Path
from typing import TYPE_CHECKING
from diot import Diot

if TYPE_CHECKING:
    from ..cmdy_bakeable import Bakeable


def register_plugins(bakeable: "Bakeable"):
    out = Diot()
    for path in Path(__file__).parent.glob("*.py"):
        if path.name.startswith("_"):
            continue

        module = import_module(f".{path.stem}", package=__package__)
        out[path.stem] = module.vendor(bakeable)

    return out



"""Utilities for cmdy"""
import inspect
import warnings
from copy import copy
from functools import wraps
from os import environ
from typing import TYPE_CHECKING, List, Tuple, Union

import curio
import executing
from diot import Diot
from simpleconf import ProfileConfig

from .cmdy_defaults import POPEN_ARG_KEYS
from .cmdy_exceptions import CmdyReturnCodeError

if TYPE_CHECKING:
    from .cmdy_result import CmdyAsyncResult


async def raise_return_code_error(aresult: "CmdyAsyncResult"):
    """Raise CmdyReturnCodeError from CmdyAsyncResult
    Compose a fake CmdyResult for CmdyReturnCodeError
    """
    # this should be
    result = Diot(
        rc=aresult._rc,
        pid=aresult.pid,
        cmd=aresult.cmd,
        piped_strcmds=getattr(aresult, "piped_strcmds", None),
        holding=Diot(okcode=aresult.holding.okcode),
        _stdout_str=(await aresult.stdout.read() if aresult.stdout else ""),
        _stderr_str=(await aresult.stderr.read() if aresult.stderr else ""),
        stdout=aresult.stdout,
        stderr=aresult.stderr,
    )

    raise CmdyReturnCodeError(result)


class SyncStreamFromAsync:
    """Take an async iterable into a sync iterable
    We use curio.run to fetch next record each time
    A StopIteration raised when a StopAsyncIteration raises
    for the async iterable
    """

    def __init__(self, astream: curio.io.FileStream, encoding: str = None):
        self.astream = astream
        self.encoding = encoding

    async def _fetch_next(self, timeout: float = None):
        await self.astream.flush()
        if timeout:
            ret = await curio.timeout_after(timeout, self.astream.__anext__)
        else:
            ret = await self.astream.__anext__()
        return ret.decode(self.encoding) if self.encoding else ret

    def next(self, timeout: float = None):
        """Fetch the next record within give timeout
        If nothing produced after the timeout, returns empty str or bytes
        """
        try:
            return curio.run(self._fetch_next(timeout))
        except StopAsyncIteration:
            raise StopIteration() from None
        except curio.TaskTimeout:
            return "" if self.encoding else b""

    def __next__(self):
        return self.next()  # pylint: disable=not-callable

    def __iter__(self):
        return self

    def dump(self):
        """Dump all records as a string or bytes"""
        return ("" if self.encoding else b"").join(self)


def property_called_as_method(caller=1):
    """Tell if a property is called by a method way"""
    frame = inspect.stack()[caller + 1].frame
    source = executing.Source.executing(frame)
    node = source.node
    try:

        return node.parent.func is node
    except AttributeError:
        return False


def property_or_method(func):
    """Make a class property also available to be called as a method"""

    @wraps(func)
    def wrapper(self):
        if property_called_as_method():

            @wraps(func)
            def func_wrapper(*args, **kwargs):
                return func(self, *args, **kwargs)

            return func_wrapper
        return func(self)

    return wrapper


def compose_arg_segment(cmd_args: dict, config: Diot) -> List[str]:
    """Compose a list of command-line arguments from the cmd_args
    by given argument composing configs, including prefix, sep and dupkey

    Note that `cmd_args` should not be reused, it will be changed in this
    function

    Examples:
        >>> compose_arg_segment({'a': 1, 'ab': 2}, {})
        >>> # ['-a', '1', '--ab', '2']
        >>> compose_arg_segment({'a': 1, 'ab': 2}, {'prefix': '--'})
        >>> # ['--a', '1', '--ab', '2']
        >>> compose_arg_segment({'a': 1, 'ab': 2}, {'sep': '='})
        >>> # ['-a=1', '--ab=2']
        >>> compose_arg_segment({'a': True, 'ab': 2}, {})
        >>> # ['-a', '--ab', '2']
        >>> compose_arg_segment({'a': False, 'ab': 2}, {})
        >>> # ['--ab', '2']
        >>> compose_arg_segment({'a': [1, 2]}, {})
        >>> # ['-a', '1', '2']
        >>> compose_arg_segment({'a': [1, 2]}, {'dupkey': True})
        >>> # ['-a', '1', '-a', '2']
        >>> compose_arg_segment({'_': [3, 4], '': [1, 2]}, {})
        >>> # ['1', '2', '3', '4']

    Args:
        cmd_args (dict): The keyword arguments for the command
        config (Diot): The configs for composing the arguments

    Returns:
        list: The composed arguments
    """
    ret: List[str] = []
    leadings = cmd_args.pop("", [])
    ret.extend(leadings if isinstance(leadings, (tuple, list)) else [leadings])

    positionals = cmd_args.pop("_", [])
    positionals = (
        positionals
        if isinstance(positionals, (tuple, list))
        else [positionals]
    )

    for key, value in cmd_args.items():
        if callable(config.deform):
            key = config.deform(key)
        prefix = (
            config.prefix
            if config.prefix != "auto"
            else "-"
            if len(key) == 1  # '' has been pop'ed
            else "--"
        )
        sep = (
            config.sep
            if config.sep != "auto"
            else " "
            if len(key) == 1
            else "="
        )
        if not isinstance(value, list):
            value = [value]

        for i, val in enumerate(value):
            if val is False:
                continue
            if sep == " ":
                if i == 0 or config.dupkey:
                    ret.append(f"{prefix}{key}")
                if val is not True:
                    ret.append(val)
            else:
                if i == 0 or config.dupkey:
                    ret.append(
                        f"{prefix}{key}"
                        + ("" if val is True else sep + str(val))
                    )
                elif val is not True:
                    ret.append(val)

    ret.extend(positionals)

    return [str(item) for item in ret]


def normalize_config(config: Diot):
    """Normalize shell and okcode to list"""
    if "okcode" in config:
        if isinstance(config.okcode, str):
            config.okcode = [okc.strip() for okc in config.okcode.split(",")]
        if not isinstance(config.okcode, list):
            config.okcode = [config.okcode]
        config.okcode = [int(okc) for okc in config.okcode]

    if "shell" in config and config.shell:
        if config.shell is True:
            config.shell = ["/bin/bash", "-c"]
        if not isinstance(config.shell, list):
            config.shell = [config.shell, "-c"]
        elif len(config.shell) == 1:
            config.shell.append("-c")


def fix_popen_config(popen_config: Diot):
    """Fix when env wrongly passed as envs.
    Send the whole `os.environ` instead of a piece of it given by
    popen_config.env
    And also raise warnings if configs should not go with popen but rather
    cmdy.
    """
    if "envs" in popen_config:
        if "env" in popen_config:
            warnings.warn(
                "Both envs and env specified in popen args, envs will"
                "be ignored"
            )
            del popen_config["envs"]
        else:
            popen_config["env"] = popen_config.pop("envs")

    if "env" in popen_config:
        normalized_env = {}
        for key, value in popen_config.env.items():
            if isinstance(value, bool):
                value = int(value)
            normalized_env[key] = str(value)

        envs = environ.copy()
        envs.update(normalized_env)

        popen_config.env = envs

    for pipe in ("stdin", "stdout", "stderr"):
        if pipe in popen_config:
            warnings.warn(
                "Motifying pipes are not allowed. " "Values will be ignored"
            )
            del popen_config[pipe]


def parse_single_kwarg(
    kwarg: dict, is_root: bool, global_config: dict
) -> "Tuple[dict, dict, dict]":
    """Parse a single kwarg that passed in `cmdy.ls({...}, ...)`
    as non-keyword arguments. Since some argument-related configurations
    can be passed in that single kwarg.
    This allows different configurations of the arguments passed to execute
    the program.

    Here we want to support the old style config item specification
    Tranditionally, we are using arguments prefixed with '_' to indicate
    that this is a config (either cmdy or popen) item rather than an argument
    that should be passed to the command.

    To do that, we first scan if there are any arguments prefixed with
    'cmdy_' or 'popen_' explictly passed. If so, we store them, and treat
    any argument starting with '_' and having the same name as an argument
    that should be passed to the command.

    For example: `{'_raise': True, 'cmdy_raise': True}` will result in
    `--_raise` as argument passing to command. To avoid that, using
    explict config name instead: `{'cmdy_raise': True}`.
    If only `{'_raise': True}` passed, it will be treated as a config item.

    A note for popen configs and cmdy configs that having the same name.
    Warnings will be give, if `encoding` and `shell` is trying to specified as
    popen configs. The reasons are:

    1. `curio.subprocess` is not handling any encoding stuff. It always
    produces bytes. `cmdy` wraps it and gives it ability to decode the bytes.
    2. `shell` is always set to `False` when the command passed to
    `curio.subprocess`. To run a command with `shell = True`, `cmdy` wraps it
    with `bash -c` or any shell that you passed to `cmdy_shell`

    Examples:
        >>> parse_single_kwarg({'a': 1, 'boy': 2, 'cmdy_prefix': '--'})
        >>> # {'a': 1, 'boy': 2}, {'prefix': '--'}, None

    Args:
        kwarg (dict): The single kwarg argument
        is_root (bool): Whether the kwarg is from the root
            (the `kwargs` from `cmdy.ls(*args, **kwargs)` directly
            For non-root kwarg, only 3 configuration items are allowed:
            `cmdy_prefix`, `cmdy_sep` and `cmdy_dupkey` and NO `popen` args
            are allowed.
        global_config (dict): The global configuration for composing the
            command-line arguments. We should have all config items
            in this global config.

    Returns:
        tuple(dict, dict, dict):
            A dict contains arguments for command only,
            local_config if not is_root else the updated global_config and,
            None if not is_root else the popen arguments

    Raises:
        ValueError: If extra arguments other than those 3 or popen
            configs are passed for non-root kwarg

    Warns:
        UserWarning: When `encoding` or `shell` passed as popen configs
    """
    # global_config = global_config.copy()

    # scan for configuration first
    local_config = Diot()
    popen_config = Diot()
    tmp_kwargs = {}
    for key, val in kwarg.items():
        if key.startswith("cmdy_"):
            local_config[key[5:]] = val
        elif key.startswith("popen_"):
            popen_config[key[6:]] = val
        else:
            tmp_kwargs[key] = val

    # kwargs without any config
    pure_cmd_kwargs = {}
    # further scan for config item starting with '_'
    for key, val in tmp_kwargs.items():
        if key.startswith("_"):
            key = key[1:]
            if key in global_config and key not in local_config:
                local_config[key] = val
            elif key in POPEN_ARG_KEYS and key not in popen_config:
                popen_config[key] = val
            else:
                pure_cmd_kwargs["_" + key] = val
        else:
            pure_cmd_kwargs[key] = val

    del tmp_kwargs

    if not is_root:
        if popen_config or any(
            key in local_config
            for key in global_config
            if key not in ("sep", "prefix", "dupkey")
        ):
            raise ValueError("Extra configs passed in argument segment.")
        return (pure_cmd_kwargs, local_config, None)

    if "shell" in popen_config:
        local_config.shell = popen_config.pop("shell")
        warnings.warn("To change the shell mode, use cmdy_shell instead.")

    if "encoding" in popen_config:
        local_config.encoding = popen_config.pop("encoding")
        warnings.warn("Please use cmdy_encoding instead of popen_encoding.")

    # global_config.update(local_config)

    return (pure_cmd_kwargs, local_config, popen_config)


def parse_args(
    name: str,
    args: tuple,
    kwargs: dict,
    # we have to pass the CMDY_CONFIG and _CMDY_BAKED_ARGS
    # here for baking purposes.
    # If we don't it will always use the original module's
    config: Diot,
    baked_args: Diot,
) -> Diot:
    """Get parse whatever passed to `cmdy.command(...)`

    Examples:
        >>> parse_args("a", "--l=a", {'x': True}, cmdy_shell=True,
                       popen_envs={...})
        >>> # ["a", "--l=a", "-x"], {"shell": True}, {"envs": {...}}

    Args:
        args (tuple): The arugments passed to `cmdy.ls(...)`
        kwargs (dict): The kwargs passed to `cmdy.ls(...)`
        cmdy_config (Config): CMDY_CONFIG from main module, passing for
            module baking purposes
        baked_args (Diot): _CMDY_BAKED_ARGS from main module, passing for
            module baking purposes

    Returns:
        tuple(list, Diot, Diot):
            The ready arguments to put to command,
            The cmdy configurations,
            The arguments will be passed to `Popen`
    """
    ret_args: list = []
    ret_kwargs: dict = {}

    # without cmdy_ prefix
    # full configs
    global_config = ProfileConfig.use_profile(config, "default", copy=True)
    # configs that only specified
    if ProfileConfig.has_profile(config, name):
        nondefault_config = ProfileConfig.use_profile(
            config,
            name,
            base=None,
            copy=True,
        )
    else:
        nondefault_config = Diot()

    ret_kwargs, baked_config, baked_popen_args = parse_single_kwarg(
        baked_args, is_root=True, global_config=global_config
    )

    global_config.update(baked_config)
    nondefault_config.update(baked_config)

    pure_cmd_kwargs, local_config, popen_config = parse_single_kwarg(
        kwargs, is_root=True, global_config=global_config
    )

    ret_kwargs.update(pure_cmd_kwargs)

    nondefault_config.update(local_config)
    global_config.update(local_config)

    baked_popen_args.update(popen_config)
    popen_config = baked_popen_args

    for arg in args:
        if isinstance(arg, dict):
            pure_cmd_kwargs_seg, local_config, _ = parse_single_kwarg(
                arg, is_root=False, global_config=global_config
            )
            if any(key in pure_cmd_kwargs for key in pure_cmd_kwargs_seg):
                warnings.warn(
                    "Argument has been specified "
                    "in both *args and **kwargs. "
                    "The one in *args will be ignored.",
                    UserWarning,
                )
                continue
            lconfig = global_config.copy()
            lconfig.update(local_config)
            ret_args.extend(compose_arg_segment(pure_cmd_kwargs_seg, lconfig))
        else:
            ret_args.append(str(arg))

    # ret_args.extend(compose_arg_segment(
    #     pure_cmd_kwargs, global_config
    # ))

    normalize_config(nondefault_config)
    fix_popen_config(popen_config)
    return Diot(
        args=ret_args,
        kwargs=ret_kwargs,
        config=nondefault_config,
        popen=popen_config,
    )


def compose_cmd(
    args: list,
    kwargs: dict,
    config: Diot,
    shell: Union[list, bool],
) -> list:
    """Compose the command for Popen.
    If shell is False, args will be directly returned. Otherwise
    args will be joined and wrapped by the shell

    Args:
        args (list): The ready args, including the executable
        kwargs (dict): The keyword arguments
        shell (str|bool): Whether we should run the whold command,
            or the path of the shell to wrap the command
            If it is True, '/bin/bash' will be used.
            Note that this is different from `Popen` behavior, which uses
            `/bin/sh` by default

    Returns:
        list: The args if shell if False, otherwise shell wrapped command
    """
    command = args[:]
    command.extend(compose_arg_segment(kwargs, config))

    if not shell:
        return command

    return shell + [" ".join(command)]  # type: ignore


def new_class(
    base: Union[type, Tuple],
    name: str = None,
    data: dict = None,
) -> type:
    """Copy the base class and copy its properties

    Subclass the base class, the property values is not copied,
    but just reference to the original one.

    Examples:
        >>> class A:
        >>>     a = [1, 2]
        >>>
        >>> class B(A):
        >>>     pass
        >>>
        >>> B.a[0] = 3
        >>> A.a  # [3, 2]
        >>>
        >>> C = copy_class(A, "C")
        >>> C.a[0] = 4
        >>> A.a  # [3, 2]
    """
    if isinstance(base, type):
        base = (base,)
    name = name or base[-1].__name__
    data = data or {}
    basedata = lambda b: {  # noqa: E731
        key: copy(val)
        for key, val in b.__dict__.items()
        if not key.startswith("__") and not callable(val)
    }
    classdata = {}
    for b in base:
        classdata.update(basedata(b))
    classdata.update(data)
    return type(name, base, classdata)



from typing import TYPE_CHECKING

from curio import subprocess

from ..cmdy_defaults import STDOUT, STDERR
from ..cmdy_exceptions import CmdyActionError

if TYPE_CHECKING:
    from ..cmdy_bakeable import Bakeable


def vendor(bakeable: "Bakeable"):
    """Vendor the plugins with the bakeable._plugin_factory"""

    @bakeable._plugin_factory.register
    class PluginValue:
        """Plugins: Value casting
        This is blocking in sync mode
        """

        @bakeable._plugin_factory.run_then
        def str(self, which=None):  # pylint: disable=redefined-builtin
            """Fetch the results as a string"""
            which = which or STDOUT
            if (
                which == STDOUT and self.holding.stdout != subprocess.PIPE
            ) or (which == STDERR and self.holding.stderr != subprocess.PIPE):
                raise CmdyActionError(
                    "Cannot fetch results from " "a redirected PIPE."
                )

            if which not in (STDOUT, STDERR):
                raise CmdyActionError("Expecting STDOUT or STDERR for which.")

            # cached results
            if (
                which == STDOUT
                and getattr(self, "_stdout_str", None) is not None
            ):
                return self._stdout_str
            if (
                which == STDERR
                and getattr(self, "_stderr_str", None) is not None
            ):
                return self._stderr_str

            self.wait()
            out = self.stdout if which == STDOUT else self.stderr
            if isinstance(out, (str, bytes)):
                setattr(
                    self,
                    "_stdout_str" if which == STDOUT else "_stderr_str",
                    out,
                )
                return out

            ret = "".join(out) if self.holding.encoding else b"".join(out)
            setattr(
                self, "_stdout_str" if which == STDOUT else "_stderr_str", ret
            )
            return ret

        @bakeable._plugin_factory.run_then
        async def astr(self, which=None):
            """Async version of str"""

            which = which or STDOUT
            if (
                which == STDOUT and self.holding.stdout != subprocess.PIPE
            ) or (which == STDERR and self.holding.stderr != subprocess.PIPE):
                raise CmdyActionError(
                    "Cannot fetch results from " "a redirected PIPE."
                )

            if which not in (STDOUT, STDERR):
                raise CmdyActionError("Expecting STDOUT or STDERR for which.")

            # cached results
            if (
                which == STDOUT
                and getattr(self, "_stdout_str", None) is not None
            ):
                return self._stdout_str
            if (
                which == STDERR
                and getattr(self, "_stderr_str", None) is not None
            ):
                return self._stderr_str

            await self.wait()
            ret = (
                (await self.stdout.read())
                if which == STDOUT
                else (await self.stderr.read())
            )

            if self.holding.encoding:
                value = ret.decode(self.holding.encoding)
                setattr(
                    self,
                    "_stdout_str" if which == STDOUT else "_stderr_str",
                    value,
                )
                return value

            setattr(
                self, "_stdout_str" if which == STDOUT else "_stderr_str", ret
            )
            return ret

        @bakeable._plugin_factory.add_method(bakeable.CmdyResult)
        def __contains__(self, item):
            return item in self.str()

        @bakeable._plugin_factory.add_method(bakeable.CmdyResult)
        def __eq__(self, other):
            return self.str() == other

        @bakeable._plugin_factory.add_method(bakeable.CmdyResult)
        def __ne__(self, other):
            return not self.__eq__(other)

        @bakeable._plugin_factory.add_method(bakeable.CmdyResult)
        def __str__(self):
            return self.str()

        @bakeable._plugin_factory.add_method(bakeable.CmdyResult)
        def __getattr__(self, name):
            if name in dir("") and not name.startswith("__"):
                return getattr(self.str(), name)
            return self.__getattribute__(name)

        @bakeable._plugin_factory.run_then
        def int(self, which=None):  # pylint: disable=redefined-builtin
            """Cast value to int"""
            return int(self.str(which))

        @bakeable._plugin_factory.run_then
        async def aint(self, which=None):
            """Async version of int"""
            return int(await self.astr(which))

        @bakeable._plugin_factory.run_then
        def float(self, which=None):  # pylint: disable=redefined-builtin
            """Cast value to float"""
            return float(self.str(which))

        @bakeable._plugin_factory.run_then
        async def afloat(self, which=None):
            """Async version of float"""
            return float(await self.astr(which))

    return PluginValue()



from typing import TYPE_CHECKING

from curio import subprocess

from ..cmdy_defaults import STDOUT, STDERR
from ..cmdy_exceptions import CmdyActionError
from ..cmdy_utils import SyncStreamFromAsync

if TYPE_CHECKING:
    from ..cmdy_bakeable import Bakeable


def vendor(bakeable: "Bakeable"):
    """Vendor the plugins with the bakeable._plugin_factory"""

    @bakeable._plugin_factory.register
    class PluginIter:
        """Plugin: iter
        Iterator over results
        """

        @bakeable._plugin_factory.add_method(bakeable.CmdyResult)
        def __iter__(self):
            return self

        @bakeable._plugin_factory.add_property(bakeable.CmdyResult)
        def stdout(self):
            """Get the iterable of stdout"""

            if self.holding.stdout != subprocess.PIPE:
                # redirected, we are unable to fetch the stdout
                return None

            if self._stdout is not None:
                return self._stdout

            orig_stdout = self._original("stdout")

            if not self.data.get("iter"):
                return orig_stdout.fget(self)

            which = self.data.iter.get("which", STDOUT)
            self._stdout = SyncStreamFromAsync(
                self.proc.stdout, self.holding.encoding
            )
            if which != STDOUT:
                self._stdout = self._stdout.dump()
            return self._stdout

        @bakeable._plugin_factory.add_property(bakeable.CmdyResult)
        def stderr(self):
            """Get the iterable of stderr"""

            if self.holding.stderr != subprocess.PIPE:
                # redirected, we are unable to fetch the stdout
                return None

            if self._stderr is not None:
                return self._stderr

            orig_stderr = self._original("stderr")

            if not self.data.get("iter"):
                return orig_stderr.fget(self)

            which = self.data.iter.get("which", STDOUT)
            self._stderr = SyncStreamFromAsync(
                self.proc.stderr, self.holding.encoding
            )
            if which != STDERR:
                self._stderr = self._stderr.dump()
            return self._stderr

        @bakeable._plugin_factory.add_method(bakeable.CmdyResult)
        def __next__(self):
            return self.next()  # pylint: disable=not-callable

        @bakeable._plugin_factory.add_method(bakeable.CmdyResult)
        def next(self, timeout=None):
            """Get next row, with a timeout limit
            If nothing produced after the timeout, returns an empty string
            """
            which = self.data.get("iter", {}).get("which", STDOUT)
            try:
                if which == STDOUT:
                    if not isinstance(self.stdout, SyncStreamFromAsync):
                        raise TypeError(
                            "CmdyResult object is not iterable "
                            "synchronously"
                        )
                    return self.stdout.next(timeout)

                if not isinstance(self.stderr, SyncStreamFromAsync):
                    raise TypeError(
                        "CmdyResult object is not iterable " "synchronously"
                    )
                return self.stderr.next(timeout)
            except StopIteration:
                # self.data.iter = {}
                self.wait()
                raise

        @bakeable._plugin_factory.run_then("it")
        def iter(self, which=None):  # pylint: disable=redefined-builtin
            """Iterator over STDOUT or STDERR of a CmdyResult object"""

            which = which or STDOUT
            self.data.iter.which = which

            if (
                which == STDOUT and self.holding.stdout != subprocess.PIPE
            ) or (which == STDERR and self.holding.stderr != subprocess.PIPE):
                raise CmdyActionError("Cannot iterate from a redirected PIPE.")

            return self

        @bakeable._plugin_factory.hold_then("it", final=True, hold_right=False)
        def iter_(self, which=None):
            """Put holding on running and iterator over STDOUT or STDERR"""
            self.should_wait = False

            if self._onhold():
                return self
            return self.run().iter(which)

    return PluginIter()



from typing import TYPE_CHECKING

from curio import subprocess

from ..cmdy_defaults import STDOUT, STDERR
from ..cmdy_exceptions import CmdyActionError

if TYPE_CHECKING:
    from ..cmdy_bakeable import Bakeable


def vendor(bakeable: "Bakeable"):
    """Vendor the plugins with the bakeable._plugin_factory"""

    @bakeable._plugin_factory.register
    class PluginPipe:
        """Plugin: pipe
        Allow piping from one command to another
        `cmdy.ls().pipe() | cmdy.cat()`
        """

        @bakeable._plugin_factory.add_property(bakeable.CmdyResult)
        def piped_cmds(self):
            """Get cmds that along the piping path

            Example:
                ```python
                c = cmdy.echo(123).p() | cmdy.cat()
                c.piped_cmds == ['echo 123', 'cat']
                ```
            """
            piped_from = self.holding.data.get("pipe", {}).get("from")
            if piped_from:
                return piped_from.piped_cmds + [self.cmd]
            return [self.cmd]

        @bakeable._plugin_factory.add_property(bakeable.CmdyResult)
        def piped_strcmds(self):
            """Get cmds that along the piping path

            Example:
                ```python
                c = cmdy.echo(123).p() | cmdy.cat()
                c.piped_cmds == ['echo 123', 'cat']
                ```
            """
            piped_from = self.holding.data.get("pipe", {}).get("from")
            if piped_from:
                return piped_from.piped_strcmds + [self.strcmd]
            return [self.strcmd]

        @bakeable._plugin_factory.add_property(bakeable.CmdyHolding)
        def piped_cmds_(self):
            """Get cmds that along the piping path

            Example:
                ```python
                c = cmdy.echo(123).p() | cmdy.cat()
                c.piped_cmds == ['echo 123', 'cat']
                ```
            """
            piped_from = self.data.get("pipe", {}).get("from")
            if piped_from:
                return piped_from.piped_cmds + [self.cmd]
            return [self.cmd]

        @bakeable._plugin_factory.add_property(bakeable.CmdyHolding)
        def piped_strcmds_(self):
            """Get cmds that along the piping path

            Example:
                ```python
                c = cmdy.echo(123).p() | cmdy.cat()
                c.piped_cmds == ['echo 123', 'cat']
                ```
            """
            piped_from = self.data.get("pipe", {}).get("from")
            if piped_from:
                return piped_from.piped_strcmds + [self.strcmd]
            return [self.strcmd]

        @bakeable._plugin_factory.add_method(bakeable.CmdyHolding)
        def __or__(
            self,
            other: bakeable.CmdyHolding,  # type: ignore
        ):

            if not self.data.get("pipe"):
                raise CmdyActionError(
                    "Piping options have been consumed or trying "
                    "to pipe from non-piping command"
                )

            assert isinstance(other, bakeable.CmdyHolding), (
                "Can only pipe to " "a CmdyHolding object."
            )

            other_pipe_data = other.data.setdefault("pipe", {})
            other_pipe_data["from"] = self

            # We shall not check the event, because the purpose here
            # is to clear the EVENT
            # But we need to check if other is also a piping command
            # which will be set if .pipe() is called
            if not other._onhold(check_event=False) and not other.data.get(
                "pipe", {}
            ).get("which"):
                self.bakeable._event.clear()
                return other.run()

            return other

        @bakeable._plugin_factory.hold_then("p")
        def pipe(self, which=None):
            """Allow command piping"""
            if self.data.get("pipe"):
                raise CmdyActionError("Unconsumed piping action.")

            # initialize data
            which = which or STDOUT
            self.data.pipe.which = which

            if (which == STDOUT and self.stdout != subprocess.PIPE) or (
                which == STDERR and self.stderr != subprocess.PIPE
            ):

                raise CmdyActionError("Cannot pipe from a redirected PIPE.")
            self.bakeable._event.set()
            return self

        @bakeable._plugin_factory.add_method(bakeable.CmdyHolding)
        def run(self, wait=None):
            """From from prior piped command"""
            orig_run = self._original("run")

            if not self.data.get("pipe", {}).get("from"):
                return orig_run(self, wait)

            prior = self.data.pipe["from"]
            prior_result = prior.run(False)
            self.data.pipe["from"] = prior

            self.stdin = (
                prior_result.proc.stdout
                if prior.data.pipe.which == STDOUT
                else prior_result.proc.stderr
            )

            ret = orig_run(self, wait)
            prior_result.wait()
            return ret

    return PluginPipe()



from shlex import quote
from typing import TYPE_CHECKING

from curio import subprocess
from diot import Diot
from varname import will

from .cmdy_defaults import get_config
from .cmdy_exceptions import CmdyExecNotFoundError, CmdyActionError
from .cmdy_utils import (
    parse_args,
    compose_arg_segment,
    compose_cmd,
    property_or_method,
    parse_single_kwarg,
)

if TYPE_CHECKING:
    from .cmdy_bakeable import Bakeable


class Cmdy:
    def __init__(self, name: str, bakeable: "Bakeable", args: Diot = None):
        self._name = name
        self._bakeable = bakeable
        self._args = args or Diot()
        self._args.setdefault("args", [])
        self._args.setdefault("kwargs", {})
        self._args.setdefault("config", {})
        self._args.setdefault("popen", {})

    def __repr__(self) -> str:
        return f"<Cmdy: {self._name} {self._args.args} @ {hex(id(self))}>"

    def __call__(self, *args, **kwargs):
        config = get_config()
        args = parse_args(
            self._name, args, kwargs, config, self._bakeable._baking_args
        )
        ready_args = self._args.args.copy() + args.args
        ready_kwargs = self._args.kwargs.copy()
        ready_kwargs.update(args.kwargs)
        ready_config = config.copy()
        ready_config.update(self._args.config)
        ready_config.update(args.config)
        ready_popen = self._args.popen.copy()
        ready_popen.update(args.popen)

        # clear direct subcommand for reuse
        self._args.args = []

        if ready_config.pop("sub", False):
            return CmdyHoldingWithSub(
                self._name,
                Diot(
                    args=ready_args,
                    kwargs=ready_kwargs,
                    config=ready_config,
                    popen=ready_popen,
                ),
                self._bakeable,
            )

        # update the executable
        exe = ready_config.pop("exe", None) or self._name
        # next attribute
        next_attr = will(raise_exc=False)

        # Let CmdyHolding handle the result
        return self._bakeable.CmdyHolding(
            Diot(
                args=[str(exe)] + ready_args,
                kwargs=ready_kwargs,
                config=ready_config,
                popen=ready_popen,
            ),
            self._bakeable,
            next_attr,
        )

    def _(self, **kwargs):
        """Bake a command"""
        if will(raise_exc=False):
            raise CmdyActionError(
                "Baking Cmdy object is supposed to be reused."
            )

        pure_cmd_kwargs, global_config, popen_config = parse_single_kwarg(
            kwargs, True, self._args.config or get_config()
        )

        kwargs = self._args.kwargs.copy() if self._args.kwargs else {}
        kwargs.update(pure_cmd_kwargs)

        config = self._args.config.copy() if self._args.config else Diot()
        config.update(global_config)

        pconfig = self._args.popen.copy() if self._args.popen else Diot()
        pconfig.update(popen_config)

        return self.__class__(
            self._name,
            self._bakeable,
            Diot(
                args=self._args.args,
                kwargs=kwargs,
                config=config,
                popen=pconfig,
            ),
        )

    def __getattr__(self, name):
        """Direct subcommand"""
        # when calling getattr(cmdy, '__wrapped__'), should return False
        if name[:2] == "__":  # pragma: no cover
            raise AttributeError

        self._args.args.append(name)
        return self


class CmdyHoldingWithSub:
    """A command with subcommands"""

    def __init__(self, name, args, bakeable):
        self._name = name
        self._args = args
        self._bakeable = bakeable

    def __getattr__(self, name):
        args = compose_arg_segment(self._args.kwargs, self._args.config)
        return Cmdy(
            self._name,
            self._bakeable,
            Diot(
                args=self._args.args + args + [name],
                kwargs={},
                config=self._args.config,
                popen=self._args.popen,
            ),
        )


class CmdyHolding:
    """Command not running yet"""

    def __new__(cls, args: Diot, bakeable: "Bakeable", will: str = None):

        holding = super().__new__(cls)

        # Use the _onhold function, but fake an object
        if cls._onhold(
            Diot(
                data=Diot(hold=False),
                bakeable=bakeable,
                did="",
                will=will,
            )
        ):
            # __init__ automatically called
            return holding

        holding.__init__(args, bakeable, will)
        result = holding.run()

        if not will:
            return result.wait()

        return result

    def __init__(self, args: Diot, bakeable: "Bakeable", will: str = None):
        # Attach the global EVENT here for later access

        # remember this for resetting
        self._reset_async = args.config["async"]
        self.bakeable = bakeable
        self.shell = args.config.shell
        self.encoding = args.config.encoding
        self.okcode = args.config.okcode
        self.timeout = args.config.timeout
        self.raise_ = args.config["raise"]
        self.should_close_fds = Diot()
        # Should I wait for the results, or just run asyncronouslly
        # This should be controlled by plugins
        # to communicate between each other
        # This only works in sync mode
        self.should_wait = False
        self.did = self.curr = ""
        self.will = will

        # pipes
        self.stdin = subprocess.PIPE
        self.stdout = subprocess.PIPE
        self.stderr = subprocess.PIPE

        args.popen.shell = False

        self.popenargs = args.popen
        # data carried by actions (ie redirect, pipe, etc)
        self.data = Diot({"async": args.config["async"], "hold": False})
        self.cmd = compose_cmd(
            args.args, args.kwargs, args.config, shell=self.shell
        )

    def __repr__(self):
        return f"<CmdyHolding: {self.cmd}>"

    def reset(self):
        """Reset the holding object for reuse"""
        # pipes
        self.stdin = subprocess.PIPE
        self.stdout = subprocess.PIPE
        self.stderr = subprocess.PIPE
        self.did = self.curr = self.will = ""

        self.should_close_fds = Diot()
        self.data = Diot({"async": self._reset_async, "hold": False})
        return self

    @property
    def strcmd(self):
        """Get the stringified cmd"""
        return " ".join(quote(cmdpart) for cmdpart in self.cmd)

    def _run(self):
        try:
            return subprocess.Popen(
                self.cmd,
                stdin=self.stdin,
                stdout=self.stdout,
                stderr=self.stderr,
                **self.popenargs,
            )
        except FileNotFoundError as fnfe:
            raise CmdyExecNotFoundError(str(fnfe)) from None

    def _onhold(self, check_event=True):
        """Tell if I am on hold
        We should be on hold to run if:
        1. EVENT is set. This means that there are unconsumed pipes
        2. The world is on hold if `.h()` or `.hold()` is called earlier
        3. If a holding-left action will be taken
        4. If a holding-right action was taken

        Args:
            check_event (bool): Should we check if event is set as well?
                We should ignore it if we are consuming a piping
        """
        return (
            (check_event and self.bakeable._event.is_set())
            or self.data.hold
            or self.will in self.bakeable._holding_left
            or self.did in self.bakeable._holding_right
        )

    @property  # type: ignore
    @property_or_method
    def async_(self):
        """Put command in async mode"""
        if self.data["async"]:
            raise CmdyActionError("Already in async mode.")

        self.data["async"] = True
        # update actions
        self.did, self.curr, self.will = (
            self.curr,
            self.will,
            will(2, raise_exc=False),
        )
        if self._onhold():
            return self
        return self.run()

    a = async_

    @property  # type: ignore
    @property_or_method
    def hold(self):
        """Put the command on hold"""
        # Whever hold is called
        self.data["hold"] = True
        self.did, self.curr, self.will = (
            self.curr,
            self.will,
            will(2, raise_exc=False),
        )

        if self.data["async"] or len(self.data) > 2:
            raise CmdyActionError(
                "Should be called in the first place: .h() or .hold()"
            )
        return self

    h = hold

    def run(self, wait=None):
        """Run the command"""
        if wait is None:
            wait = self.should_wait
        if not self.data["async"]:
            ret = self.bakeable.CmdyResult(self._run(), self)
            if wait:
                return ret.wait()
            return ret

        return self.bakeable.CmdyAsyncResult(self._run(), self)



import inspect
from shlex import quote

import curio
from curio import subprocess
from diot import Diot

from .cmdy_defaults import STDOUT
from .cmdy_exceptions import CmdyTimeoutError, CmdyReturnCodeError
from .cmdy_utils import SyncStreamFromAsync, raise_return_code_error


class CmdyResult:

    """Sync version of result"""

    def __init__(self, proc, holding):
        self.proc = proc
        self.holding = holding
        self.did = self.curr = ""
        self.will = holding.will
        self._stdout = None
        self._stderr = None
        self.data = Diot()
        self._rc = None

    def __repr__(self):
        return f"<CmdyResult: {self.cmd}>"

    @property
    def rc(self):
        """Get the return code"""
        if self._rc is not None:
            return self._rc
        self.wait()
        return self._rc

    @property
    def pid(self):
        """Get the pid of the process"""
        return self.proc.pid

    @property
    def cmd(self):
        """Get the stringified command"""
        return self.holding.cmd

    @property
    def strcmd(self):
        """Get the stringified cmd"""
        return " ".join(quote(cmdpart) for cmdpart in self.cmd)

    def wait(self):
        """Wait until command is done"""
        timeout = self.holding.timeout
        try:
            if timeout:
                self._rc = curio.run(
                    curio.timeout_after(timeout, self.proc.wait)
                )
            else:
                self._rc = curio.run(self.proc.wait())
        except curio.TaskTimeout:
            self.proc.kill()
            raise CmdyTimeoutError(
                f"Timeout after {self.holding.timeout} seconds."
            ) from None
        else:
            if self._rc not in self.holding.okcode and self.holding.raise_:
                raise CmdyReturnCodeError(self)
            return self
        finally:
            self._close_fds()

    def _close_fds(self):
        if not self.holding.should_close_fds:
            return
        for filed in self.holding.should_close_fds.values():
            if filed:
                filed.close()

    @property
    def stdout(self):
        """The stdout of the command"""
        if self.holding.stdout != subprocess.PIPE:
            # redirected, we are unable to fetch the stdout
            return None

        if self._stdout is not None:
            return self._stdout

        self._stdout = SyncStreamFromAsync(
            self.proc.stdout, encoding=self.holding.encoding
        ).dump()
        return self._stdout

    @property
    def stderr(self):
        """The stderr of the command"""
        if self.holding.stderr != subprocess.PIPE:
            # redirected, we are unable to fetch the stdout
            return None

        if self._stderr is not None:
            return self._stderr
        self._stderr = SyncStreamFromAsync(
            self.proc.stderr, encoding=self.holding.encoding
        ).dump()
        return self._stderr


class CmdyAsyncResult:
    """Asyncronous result"""

    def __repr__(self):
        return f"<CmdyAsyncResult: {self.cmd}>"

    async def _close_fds(self):
        if not self.holding.should_close_fds:
            return
        try:
            for filed in self.holding.should_close_fds.values():
                if filed:
                    coro = filed.close()
                    if inspect.iscoroutine(coro):
                        await coro  # pragma: no cover
        except AttributeError:  # pragma: no cover
            pass

    def __aiter__(self):
        return self

    async def __anext__(self):
        which = self.data.get("iter", {}).get("which", STDOUT)
        stream = self.stdout if which == STDOUT else self.stderr
        try:
            line = await stream.__anext__()
        except StopAsyncIteration:
            await self.wait()
            raise
        if self.holding.encoding:
            line = line.decode(self.holding.encoding)
        return line

    async def wait(self):
        timeout = self.holding.timeout

        try:
            if timeout:
                self._rc = await curio.timeout_after(timeout, self.proc.wait)
            else:
                self._rc = await self.proc.wait()
        except curio.TaskTimeout:
            self.proc.kill()
            raise CmdyTimeoutError(
                "Timeout after " f"{self.holding.timeout} seconds."
            ) from None
        else:
            if self._rc not in self.holding.okcode and self.holding.raise_:
                await raise_return_code_error(self)
            return self
        finally:
            await self._close_fds()

    @property
    async def rc(self):
        if self._rc is not None:
            return self._rc
        await self.wait()
        return self._rc

    @property
    def stdout(self):
        return self.proc.stdout

    @property
    def stderr(self):
        return self.proc.stderr



from typing import TYPE_CHECKING, Any

from ..cmdy_defaults import STDIN, STDOUT, STDERR
from ..cmdy_exceptions import CmdyActionError

if TYPE_CHECKING:
    from ..cmdy_bakeable import Bakeable


def vendor(bakeable: "Bakeable"):
    """Vendor the plugins with the bakeable._plugin_factory"""

    @bakeable._plugin_factory.register
    class PluginRedirect:
        """Plugin: redirect
        Redirect the in/out to somewhere else"""

        def _redirect(
            self: bakeable.CmdyHolding,   # type: ignore
            which: list,
            append: bool,
            file: Any,
        ) -> bool:
            # add file-like type suport for file
            if not self.data.get("redirect"):
                raise CmdyActionError(
                    "Cannot redirect a non-redirecting command. "
                    "Did you forget to call .r(), .redir() or .redirect()?"
                )
            curr_pipe = which.pop(0)
            self.data.redirect.which = which

            if curr_pipe == STDIN:
                if isinstance(file, bakeable.CmdyResult):
                    self.stdin = file.proc.stdout
                    self.should_close_fds.stdin = None
                elif hasattr(file, "read"):
                    self.stdin = file
                    self.should_close_fds.stdin = None
                else:
                    self.stdin = open(file, "r", encoding=self.encoding)
                    self.should_close_fds.stdin = self.stdin

            elif curr_pipe == STDOUT:
                if file == STDERR:
                    raise CmdyActionError("Cannot redirect STDOUT to STDERR.")
                if hasattr(file, "read"):
                    self.stdout = file
                    self.should_close_fds.stdout = None
                else:
                    self.stdout = open(
                        file, "a" if append else "w", encoding=self.encoding
                    )
                    self.should_close_fds.stdout = self.stdout
            elif curr_pipe == STDERR:
                if file == STDOUT:
                    self.stderr = STDOUT
                    self.should_close_fds.stderr = None
                elif hasattr(file, "read"):
                    self.stderr = file
                    self.should_close_fds.stderr = None
                else:
                    self.stderr = open(
                        file, "a" if append else "w", encoding=self.encoding
                    )
                    self.should_close_fds.stderr = self.stderr
            else:
                raise CmdyActionError(
                    "Don't know what to redirect. "
                    "Expecting STDIN, STDOUT or STDERR"
                )

            # Since we are holding right, set did to ''
            # to let the right action run
            self.did = ""
            if not which and not self._onhold():
                # self.data.redirect = {}
                return self.run()

            return self

        @bakeable._plugin_factory.add_method(bakeable.CmdyHolding)
        def __gt__(self, file):
            which = self.data.get("redirect", {}).get("which", [STDOUT])
            return PluginRedirect._redirect(self, list(which), False, file)

        @bakeable._plugin_factory.add_method(bakeable.CmdyHolding)
        def __lt__(self, file):
            which = self.data.get("redirect", {}).get("which", [STDOUT])
            return PluginRedirect._redirect(self, list(which), False, file)

        @bakeable._plugin_factory.add_method(bakeable.CmdyHolding)
        def __xor__(self, file):
            """Priority issue with gt (>)
            We need brackets to ensure the order:
            `(cmdy.ls().r(REDIR_RSPT) > outfile) > errfile`
            To avoid this, use xor (^) instead
            """
            which = self.data.get("redirect", {}).get("which", [STDOUT])
            if which[0] == STDIN:
                return self.__lt__(file)
            return self.__gt__(file)

        @bakeable._plugin_factory.add_method(bakeable.CmdyHolding)
        def __rshift__(self, file):
            which = self.data.get("redirect", {}).get("which", [STDOUT])
            return PluginRedirect._redirect(self, list(which), True, file)

        @bakeable._plugin_factory.hold_then("r,redir", hold_right=True)
        def redirect(self, *which):
            """Redirect the input/output"""

            # We should wait for the command to finish, so that we
            # don't leave it piping in background
            # To do so, use it in async mode
            self.should_wait = True

            which = which or [STDOUT]

            # since this is final, so
            # cmdy.ls().r().r() will never happen
            if self.data.redirect:  # pragma: no cover
                raise CmdyActionError("Unconsumed redirect action.")

            # initialize data
            self.data.redirect.which = list(which)

            return self

    return PluginRedirect()



from functools import wraps
from typing import Callable

from varname import will

from .cmdy_exceptions import CmdyActionError
from .cmdy_utils import property_or_method


def _method_enable(cls, names, func):
    for name in names:
        # put original func into stack if any
        stack = cls._plugin_stacks.setdefault(name, [])
        orig_func = getattr(cls, name, None)

        if orig_func:
            stack.insert(0, orig_func)
        setattr(cls, name, func)


def _method_disable(cls, names, func):
    for name in names:
        # remove the function from stack
        # and restore the latest defined one
        curr_func = getattr(cls, name)
        if curr_func is func:
            delattr(cls, name)

        if func in cls._plugin_stacks[name]:
            cls._plugin_stacks[name].remove(func)

        if not hasattr(cls, name) and cls._plugin_stacks[name]:
            setattr(cls, name, cls._plugin_stacks[name].pop(0))


def _plugin_funcname(func):
    """Get the function name defined in a plugin
    We will ignore the underscores on the right except for those
    magic method, so that we can have the same function defined for
    different classes
    """
    funcname = func.__name__.rstrip("_")
    if funcname.startswith("__"):
        return funcname + "__"
    return funcname


def _add_method(cls: type) -> Callable:
    """A decorator to add a method to a class"""

    def decorator(func):
        func.enable = lambda: _method_enable(
            cls, [_plugin_funcname(func)], func
        )
        func.disable = lambda: _method_disable(
            cls, [_plugin_funcname(func)], func
        )
        return func

    return decorator


def _add_property(cls):
    """A decorator to add a property to a class"""

    def decorator(func):
        func.enable = lambda: _property_enable(
            cls, [_plugin_funcname(func)], func
        )
        func.disable = lambda: _property_disable(
            cls, [_plugin_funcname(func)], func
        )
        return func

    return decorator


def _property_enable(cls, names, func):
    for name in names:
        stack = cls._plugin_stacks.setdefault(name, [])
        orig_prop = getattr(cls, name, None)
        if orig_prop:
            stack.insert(0, orig_prop)
        setattr(cls, name, property(func))


def _property_disable(cls, names, func):
    for name in names:
        curr_prop = getattr(cls, name)
        if curr_prop.fget is func:
            delattr(cls, name)

        cls._plugin_stacks[name] = [
            prop for prop in cls._plugin_stacks[name] if prop.fget is not func
        ]

        if not hasattr(cls, name) and cls._plugin_stacks[name]:
            setattr(cls, name, cls._plugin_stacks[name].pop(0))


def pluginable(cls):
    """Put hooks into the original class for extending"""
    # store the functions with the same name
    # that defined by different plugins
    # Note that current (most recently added) is not in the stack
    cls._plugin_stacks = {}

    def _original(self, fname):
        """Get the original function of self, if it is overridden"""
        # callframe is oringally -1
        frame = self._plugin_callframe.setdefault(fname, -1)
        frame += 1
        self._plugin_callframe[fname] = frame
        # print(cls._plugin_stacks)
        return cls._plugin_stacks[fname][frame]

    cls._original = _original

    orig_init = cls.__init__

    def __init__(self, *args, **kwargs):
        self._plugin_callframe = {}
        orig_init(self, *args, **kwargs)

    cls.__init__ = __init__

    if cls.__name__ == "CmdyHolding":
        orig_reset = cls.reset

        @wraps(orig_reset)
        def reset(self, *args, **kwargs):
            # clear the callframes as well
            self._plugin_callframe = {}
            orig_reset(self, *args, **kwargs)
            return self

        cls.reset = reset

    return cls


class PluginFactory:
    def __init__(
        self,
        bakeable,
    ):
        self.bakeable = bakeable

    def register(self, plugin: type):
        """Register a plugin"""
        orig_init = plugin.__init__  # type: ignore
        data = [
            val for val in plugin.__dict__.values() if hasattr(val, "enable")
        ]

        @wraps(orig_init)
        def __init__(self):
            self.enabled = False
            self.enable()
            orig_init(self)

        def enable(self):
            """Enable all the functions properties defined in the plugin"""
            for val in data:
                val.enable()
            self.enabled = True

        def disable(self):
            """Disable all the functions properties defined in the plugin"""
            for val in data:
                val.disable()
            self.enabled = False

        plugin.enable = enable
        plugin.disable = disable
        plugin.__init__ = __init__  # type: ignore
        return plugin

    def add_method(self, cls: type):
        """Get the add method hook"""
        return _add_method(cls)

    def add_property(self, cls: type):
        """Get the add property hook"""
        return _add_property(cls)

    def _plugin_then(
        self,
        cls,
        func,
        aliases,
        final: bool = False,
        prop: bool = True,
        hold_right: bool = False,
    ) -> Callable:

        aliases = aliases or []
        if not isinstance(aliases, list):
            aliases = [alias.strip() for alias in aliases.split(",")]
        aliases.insert(0, _plugin_funcname(func))

        finals = (
            self.bakeable._holding_finals
            if cls is self.bakeable.CmdyHolding
            else self.bakeable._result_finals
        )

        if final:
            finals.extend(aliases)

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            # Update actions
            self.did, self.curr = self.curr, self.will
            self.will = will(2, raise_exc=False)

            if (
                self.curr in finals
                and self.will in self.bakeable._holding_left
            ):
                raise CmdyActionError("Action taken after a final action.")
            # Initialize data
            # Make it True to tell future actions that I have been called
            # Just in case some plugins forget to do this
            self.data.setdefault(_plugin_funcname(func), {})
            return func(self, *args, **kwargs)

        if prop:
            wrapper.enable = lambda: _property_enable(
                cls, aliases, property_or_method(wrapper)
            )
            wrapper.disable = lambda: _property_disable(
                cls, aliases, property_or_method(wrapper)
            )
        else:
            wrapper.enable = lambda: _method_enable(cls, aliases, wrapper)
            wrapper.disable = lambda: _method_disable(cls, aliases, wrapper)

        if cls is self.bakeable.CmdyHolding:
            self.bakeable._holding_left.extend(aliases)
            if hold_right:
                self.bakeable._holding_right.extend(aliases)

        return wrapper

    def hold_then(
        self,
        alias_or_func=None,
        final: bool = False,
        prop: bool = True,
        hold_right: bool = True,
    ) -> Callable:
        """What to do if a command is holding

        Args:
            alias_or_func (str|list|Callable): Direct decorator or with kwargs
            final (bool): If this is a final action
            prop (bool): Enable the property call for the method
                Only works for CmdyHolding
            hold_right (bool): Tell previous actions that I should be on hold.
                            But make sure running will be taken good care of.
        """

        aliases, func = (
            (None, alias_or_func)
            if callable(alias_or_func)
            else (alias_or_func, None)
        )

        if func:
            return self._plugin_then(
                self.bakeable.CmdyHolding,
                func,
                aliases,
                final=final,
                prop=prop,
                hold_right=hold_right,
            )

        return lambda func: self._plugin_then(
            self.bakeable.CmdyHolding,
            func,
            aliases,
            final=final,
            prop=prop,
            hold_right=hold_right,
        )

    def run_then(
        self,
        alias_or_func=None,
        final: bool = False,
    ) -> Callable:
        """What to do if a command is running"""
        aliases, func = (
            (None, alias_or_func)
            if callable(alias_or_func)
            else (alias_or_func, None)
        )

        if func:
            return self._plugin_then(
                self.bakeable.CmdyResult,
                func,
                aliases,
                final=final,
                prop=False,
            )

        return lambda func: self._plugin_then(
            self.bakeable.CmdyResult,
            func,
            aliases,
            final=final,
            prop=False,
        )



from threading import Event

from .cmdy_plugin import PluginFactory
from .cmdy_plugins import register_plugins
from .cmdy_exceptions import (
    CmdyActionError,
    CmdyTimeoutError,
    CmdyExecNotFoundError,
    CmdyReturnCodeError,
)
from .cmdy_defaults import STDIN, STDOUT, STDERR, DEVNULL
from .cmdy_plugin import pluginable
from .cmdy_result import CmdyResult, CmdyAsyncResult
from .cmdy_utils import new_class
from .cmdy import Cmdy, CmdyHolding


class Bakeable:
    def __init__(self, **baking_args):
        self.CmdyActionError = CmdyActionError
        self.CmdyTimeoutError = CmdyTimeoutError
        self.CmdyExecNotFoundError = CmdyExecNotFoundError
        self.CmdyReturnCodeError = CmdyReturnCodeError
        self.CmdyResult = pluginable(
            new_class(CmdyResult, data={"__module__": "cmdy"})
        )
        self.CmdyHolding = pluginable(
            new_class(CmdyHolding, data={"__module__": "cmdy"})
        )
        self.Cmdy = Cmdy
        self.STDIN = STDIN
        self.STDOUT = STDOUT
        self.STDERR = STDERR
        self.DEVNULL = DEVNULL
        self._event = Event()
        self._baking_args = baking_args
        self._holding_left = ["a", "async_", "h", "hold"]
        self._holding_right = []
        self._holding_finals = []
        self._result_finals = []
        # init plugins
        self._plugin_factory = PluginFactory(self)
        self._plugins = register_plugins(self)
        self.CmdyAsyncResult = new_class(
            self.CmdyResult,
            "CmdyAsyncResult",
            {"__module__": "cmdy", **CmdyAsyncResult.__dict__},
        )

    def __call__(self, **baking_args):
        return self.__class__(**baking_args)

    def __getattr__(self, name: str):
        return self.Cmdy(name, bakeable=self)



import inspect
from os import devnull
from functools import lru_cache
from subprocess import Popen

from diot import Diot
from simpleconf import ProfileConfig

_DEFAULT_CONFIG = Diot(
    {
        "async": False,
        "deform": lambda name: name.replace("_", "-"),
        "dupkey": False,
        "exe": None,
        "encoding": "utf-8",
        "okcode": [0],
        "prefix": "auto",
        "raise": True,
        "sep": " ",
        "shell": False,
        "sub": False,
        "timeout": 0,
    }
)

STDIN = -7
STDOUT = -2
STDERR = -8
DEVNULL = devnull

# Sometimes we may occasionally use envs instead env
POPEN_ARG_KEYS = inspect.getfullargspec(Popen).args + ["envs"]


@lru_cache()
def get_config() -> Diot:
    return ProfileConfig.load(
        {"default": _DEFAULT_CONFIG},
        "~/.cmdy.toml",
        "./.cmdy.toml",
        "CMDY.osenv",
        ignore_nonexist=True,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cmdy', 'cmdy.cmdy_plugins']

package_data = \
{'': ['*']}

install_requires = \
['curio>=1.5,<2.0',
 'diot>=0.1,<0.2',
 'python-simpleconf[toml]>=0.5,<0.6',
 'varname>=0.8,<0.9']

setup_kwargs = {
    'name': 'cmdy',
    'version': '0.5.0',
    'description': 'Shell language to run command in python',
    'long_description': '# cmdy\n"Shell language" to run command in python\n\n[![pypi][1]][2] [![tag][3]][4] [![travis][5]][6] [![codacy quality][7]][8] [![codacy quality][9]][8] ![pyver][10]\n\n## Installation\n```shell\npip install cmdy\n```\n\n## Usage\n\nSee [Demo](./demo.ipynb)\n\n### Basic usage\n```python\nfrom cmdy import ls\nprint(ls())\n```\n\n```python\nfor line in ls().iter():\n    print(\'Got:\', line, end=\'\')\n```\n\n#### With non-keyword arguments\n```python\nfrom cmdy import tar\nprint(tar("cvf", "/tmp/test.tar", "./cmdy"))\n```\n\n#### With keyword arguments\n```python\nfrom cmdy import curl\ncurl("http://duckduckgo.com/", o="/tmp/page.html", silent=True)\n# curl http://duckduckgo.com/ -o /tmp/page.html --silent\n```\n\n#### Order keyword arguments\n```python\ncurl("http://duckduckgo.com/", "-o", "/tmp/page.html", "--silent")\n# or\n\nfrom diot import OrderedDiot\nkwargs = OrderedDiot()\nkwargs.silent = True\nkwargs.o = \'/tmp/page.html\'\ncurl("http://duckduckgo.com/", kwargs)\n# You can also use collections.OrderedDict\n```\n\n#### Prefix and separator for keyword arguments\n```python\nfrom cmdy import bedtools, bcftools\nbedtools.intersect(wa=True, wb=True,\n                   a=\'query.bed\', b=[\'d1.bed\', \'d2.bed\', \'d3.bed\'],\n                   names=[\'d1\', \'d2\', \'d3\'], sorted=True,\n                   _prefix=\'-\').h().strcmd\n# \'bedtools intersect -wa -wb -a query.bed \\\n# -b d1.bed d2.bed d3.bed -names d1 d2 d3 -sorted\'\n```\n\n```python\n# default prefix is auto\nbcftools.query(_=[\'a.vcf\', \'b.vcf\'], H=True,\n               format=\'%CHROM\\t%POS\\t%REF\\t%ALT\\n\').h().strcmd\n\n# "bcftools query -H --format \'%CHROM\\t%POS\\t%REF\\t%ALT\\n\' a.vcf b.vcf"\n\nls(l=True, block_size=\'KB\', _sep=\'auto\').h().cmd\n[\'ls\', \'-l\', \'--block-size=KB\']\n```\n\n#### Mixed combinations of prefices and separators in one command\n```python\nfrom cmdy import java\n# Note this is just an example for old verion picard.\n# Picard is changing it\'s style\n\npicard = java(jar=\'picard.jar\', _prefix=\'\', _sep=\'=\', _sub=True)\nc = picard.SortSam(I=\'input.bam\', O=\'sorted.bam\',\n               SORTED_ORDER=\'coordinate\',\n               _prefix=\'\', _sep=\'=\', _deform=None).h\nprint(c.cmd)\n\n# same as the above\njava({\'jar\': \'picard.jar\', \'_prefix\': \'-\', \'_sep\': \' \'},\n     \'SortSam\', I=\'input.bam\', O=\'sorted.bam\',\n     SORTED_ORDER=\'coordinate\', _prefix=\'\', _sep=\'=\', _deform=None).h().cmd\n\n# _deform prevents SORTED_ORDER to be deformed to SORTED-ORDER\n\n# [\'java\', \'jar=picard.jar\',\n#  \'SortSam\', \'I=input.bam\', \'O=sorted.bam\', \'SORTED_ORDER=coordinate\']\n```\n\n#### Subcommands\n\n```python\nfrom cmdy import git\ngit.branch(v=True).fg\n# <CmdyResult: [\'git\', \'branch\', \'-v\']>\n```\n\n```python\n# What if I have separate arguments for main and sub-command?\ngit(git_dir=\'.\', _sub=True).branch(v=True).h\n# <CmdyHolding: [\'git\', \'--git-dir\', \'.\', \'branch\', \'-v\']>\n```\n\n#### Duplicated keys for list arguments:\n```python\nfrom cmdy import sort\nprint(sort(k=[\'1,1\', \'2,2\'], t=\'_\', _=\'./.editorconfig\', _dupkey=True))\n# sort -k 1,1 -k 2,2 ./.editorconfig\n```\n\n### Return code and exception\n```python\nfrom cmdy import x\nx()\n```\n\n```python console\nTraceback (most recent call last):\n  File "<ipython-input-16-092cc5b72e61>", line 2, in <module>\n    x()\n/path/.../to/cmdy/__init__.py", line 146, in __call__\n    ready_cfgargs, ready_popenargs, _will())\n/path/.../to/cmdy/__init__.py", line 201, in __new__\n    result = holding.run()\n/path/.../to/cmdy/__init__.py", line 854, in run\n    return orig_run(self, wait)\n/path/.../to/cmdy/__init__.py", line 717, in run\n    return orig_run(self, wait)\n/path/.../to/cmdy/__init__.py", line 327, in run\n    ret = CmdyResult(self._run(), self)\n/path/.../to/cmdy/__init__.py", line 271, in _run\n    raise CmdyExecNotFoundError(str(fnfe)) from None\ncmdy.cmdy_util.CmdyExecNotFoundError: [Errno 2] No such file or directory: \'x\': \'x\'\n```\n\n```python\nfrom cmdy import ls\nls(\'non-existing-file\')\n```\n\n```python console\n\nTraceback (most recent call last):\n  File "<ipython-input-17-132683fc2227>", line 2, in <module>\n    ls(\'non-existing-file\')\n/path/.../to/cmdy/__init__.py", line 146, in __call__\n    ready_cfgargs, ready_popenargs, _will())\n/path/.../to/cmdy/__init__.py", line 204, in __new__\n    return result.wait()\n/path/.../to/cmdy/__init__.py", line 407, in wait\n    raise CmdyReturnCodeError(self)\ncmdy.cmdy_util.CmdyReturnCodeError: Unexpected RETURN CODE 2, expecting: [0]\n\n  [   PID] 167164\n\n  [   CMD] [\'ls non-existing-file\']\n\n  [STDOUT]\n\n  [STDERR] ls: cannot access non-existing-file: No such file or directory\n```\n\n#### Don\'t raise exception but store the return code\n```python\nfrom cmdy import ls\nresult = ls(\'non-existing-file\', _raise=False)\nresult.rc # 2\n```\n\n#### Tolerance on return code\n```python\nfrom cmdy import ls\n# or _okcode=[0,2]\nls(\'non-existing-file\', _okcode=\'0,2\').rc # 2\n```\n\n### Timeouts\n```python\nfrom cmdy import sleep\nsleep(3, _timeout=1)\n```\n\n```python console\nTraceback (most recent call last):\n  File "<ipython-input-20-47b0ec7af55f>", line 2, in <module>\n    sleep(3, _timeout=1)\n/path/.../to/cmdy/__init__.py", line 146, in __call__\n    ready_cfgargs, ready_popenargs, _will())\n/path/.../to/cmdy/__init__.py", line 204, in __new__\n    return result.wait()\n/path/.../to/cmdy/__init__.py", line 404, in wait\n    ) from None\ncmdy.cmdy_util.CmdyTimeoutError: Timeout after 1 seconds.\n```\n\n### Redirections\n```python\nfrom cmdy import cat\ncat(\'./pytest.ini\').redirect > \'/tmp/pytest.ini\'\nprint(cat(\'/tmp/pytest.ini\'))\n```\n\n#### Appending\n```python\n# r short for redirect\ncat(\'./pytest.ini\').r >> \'/tmp/pytest.ini\'\nprint(cat(\'/tmp/pytest.ini\')) # content doubled\n```\n\n#### Redirecting to a file handler\n```python\nwith open(\'/tmp/pytest.ini\', \'w\') as f\n    cat(\'./pytest.ini\').r > f\n\nprint(cat(\'/tmp/pytest.ini\'))\n```\n\n#### STDIN, STDOUT and/or STDERR redirections\n```python\nfrom cmdy import STDIN, STDOUT, STDERR, DEVNULL\n\nc = cat().r(STDIN) < \'/tmp/pytest.ini\'\nprint(c)\n```\n\n```python\n# Mixed\nc = cat().r(STDIN, STDOUT) ^ \'/tmp/pytest.ini\' > DEVNULL\n# we can\'t fetch result from a redirected pipe\nprint(c.stdout)\n\n# Why not \'<\' for STDIN?\n# Because the priority of the operator is not in sequential order.\n# We can use < for STDIN, but we need to ensure it runs first\nc = (cat().r(STDIN, STDOUT) < \'/tmp/pytest.ini\') > DEVNULL\nprint(c.stdout)\n\n# A simple rule for multiple redirections to always use ">" in the last place\n```\n\n```python\n# Redirect stderr to stdout\nfrom cmdy import bash\nc = bash(c="cat 1>&2").r(STDIN, STDERR) ^ \'/tmp/pytest.ini\' > STDOUT\nprint(c.stdout)\n```\n\n```python\n# Redirect the world\nc = bash(c="cat 1>&2").r(STDIN, STDOUT, STDERR) ^ \'/tmp/pytest.ini\' ^ DEVNULL > STDOUT\nprint(c.stdout) # None\nprint(c.stderr) # None\n```\n\n### Pipings\n```python\nfrom cmdy import grep\nc = ls().p | grep(\'README\')\nprint(c)\n# README.md\n# README.rst\n```\n\n```python\n# p short for pipe\nc = ls().p | grep(\'README\').p | grep(\'md\')\nprint(c) # README.md\nprint(c.piped_strcmds) # [\'ls\', \'grep README\', \'grep md\']\n```\n\n```python\nfrom cmdy import _CMDY_EVENT\n# !!! Pipings should be consumed immediately!\n# !!! DO NOT do this\nls().p\nls() # <- Will not run as expected\n# All commands will be locked as holding until pipings are consumed\n_CMDY_EVENT.clear()\nprint(ls()) # runs\n\n# See Advanced/Holdings if you want to hold a piping command for a while\n```\n\n### Running command in foreground\n```python\nls().fg\n```\n\n```python\nfrom cmdy import tail\ntail(\'/tmp/pytest.ini\', f=True, _timeout=3).fg\n# This mimics the `tail -f` program\n# You will see the content comes out one after another\n# and then program hangs for 3s\n```\n\nYou can also write an `echo-like` program easily. See \'[echo.py](./echo.py)\'\n\n### Iterating on output\n```python\nfor line in ls().iter():\n    print(line, end=\'\')\n```\n\n#### Iterating on stderr\n```python\nfor line in bash(c="cat /tmp/pytest.ini 1>&2").iter(STDERR):\n    print(line, end=\'\')\n```\n\n#### Getting live output\n```python\n# Like we did for `tail -f` program\n# This time, we can do something with each output line\n\n# Let\'s use a thread to write content to a file\n# And we try to get the live contents using cmdy\nimport time\nfrom threading import Thread\ndef live_write(file, n):\n\n    with open(file, \'w\', buffering=1) as f:\n        # Let\'s write something every half second\n        for i in range(n):\n            f.write(str(i) + \'\\n\')\n            time.sleep(.5)\n\ntest_file = \'/tmp/tail-f.txt\'\nThread(target=live_write, args=(test_file, 10)).start()\n\nfrom cmdy import tail\n\ntail_iter = tail(f=True, _=test_file).iter()\n\nfor line in tail_iter:\n    # Do whatever you want with the line\n    print(\'We got:\', line, end=\'\')\n    if line.strip() == \'8\':\n        break\n\n# make sure thread ends\ntime.sleep(2)\n```\n\n```python\n# What about timeout?\n\n# Of course you can use a timer to check inside the loop\n# You can also set a timeout for each fetch\n\n# Terminate after 10 queries\n\nThread(target=live_write, args=(test_file, 10)).start()\n\nfrom cmdy import tail\n\ntail_iter = tail(f=True, _=test_file).iter()\n\nfor i in range(10):\n    print(\'We got:\', tail_iter.next(timeout=1), end=\'\')\n```\n\n### Advanced\n#### Baking the `cmdy` object\n\nSometimes, you may want to run the same program a couple of times, with the same set of arguments or configurations, and you don\'t want to type those arguments every time, then you can bake the Cmdy object with that same arguments or configurations.\n\nFor example, if you want to run ls as ls -l all the time:\n\n```python\nfrom cmdy import ls\nll = ls._(l=True)\nprint(ll().h.cmd) # [\'ls\', \'-l\']\nprint(ll(a=True).h.cmd) # [\'ls\', \'-l\', \'-a\']\n# I don\'t want the l flag for some commands occasionally\nprint(ll(l=False).h.cmd) # [\'ls\']\n\n# Bake a baked command\nlla = ll._(a=True)\nprint(lla().h.cmd) # [\'ls\', \'-l\', \'-a\']\n```\n\n```python\n# I know git is always gonna run with subcommand\ngit = git._(_sub=True)\n# don\'t bother to pass _sub=True every time\nprint(git(git_dir=\'.\').branch(v=True).h)\n# <CmdyHolding: [\'git\', \'--git-dir\', \'.\', \'branch\', \'-v\']>\nprint(git().status().h)\n# <CmdyHolding: [\'git\', \'status\']>\n```\n\n```python\n# What if I have a subcommand call bake?\nfrom cmdy import git, CmdyActionError\n\nprint(git.branch().h.cmd) # [\'git\', \'branch\']\nprint(type(git._())) # <class \'cmdy.Cmdy\'>\n\n# run the git with _sub\nprint(git(_sub=True).bake().h.cmd) # [\'git\', \'bake\']\n```\n\n#### Baking the whole module\n```python\nimport cmdy\n# run version of the whole world\nsh = cmdy(version=True)\n# anything under sh directly will be supposed to have subcommand\nfrom sh import git, gcc\nprint(git().h)\n# <CmdyHolding: [\'git\', \'--version\']>\nprint(gcc().h)\n# <CmdyHolding: [\'gcc\', \'--version\']>\n```\n\nNote that module baking is deep copying, except the exception classes and some utils. This means, you would expect following behavior:\n\n```python\nimport cmdy\nfrom cmdy import CmdyHolding, CmdyExecNotFoundError\n\nsh = cmdy()\n\nc = sh.echo().h\nprint(type(c)) # <class \'cmdy.CmdyHolding\'>\nprint(isinstance(c, CmdyHolding)) # False\nprint(isinstance(c, sh.CmdyHolding)) # True\n\ntry:\n    sh.notexisting()\nexcept CmdyExecNotFoundError:\n    # we can catch it, as CmdyExecNotFoundError is sh.CmdyExecNotFoundError\n    print(\'Catched!\')\n```\n\n#### Holding objects\n\nYou may have noticed that we have a couple of examples above with a final call .h or .h(), which is holding the command from running.\n\nYou can do that, too, if you have multiple operations\n\n```python\nprint(ls().h) # <CmdyHolding: [\'ls\']>\n\n# however, you cannot hold after some actions\nls().r.h\n# CmdyActionError: Should be called in the first place: .h() or .hold()\n```\n\nOnce a command is on hold (by .h, .hold, .h() or .hold())\n\nYou have to explictly call run() to set the command running\n\n```python\nfrom time import time\ntic = time()\nc = sleep(2).h\nprint(f\'Time elapsed: {time() - tic:.3f} s\')\n# Time elapsed: 0.022 s\n\n# not running even with fg\nc.fg\nprint(f\'Time elapsed: {time() - tic:.3f} s\')\n# Time elapsed: 0.034 s\nc.run()\nprint(f\'Time elapsed: {time() - tic:.3f} s\')\n# Time elapsed: 2.043 s\n```\n\n#### Reuse of command\n```python\n# After you set a command running,\n# you can retrieve the holding object,\n# and reuse it\nfrom cmdy import ls\nc = ls().fg\n# nothing will be produced\nc.holding.reset().r > DEVNULL\n```\n\n#### Async mode\n```python\nimport curio\nfrom cmdy import ls\na = ls().a # async command is never blocking!\n\nasync def main():\n    async for line in a:\n        print(line, end=\'\')\n\ncurio.run(main())\n```\n\n#### Extending `cmdy`\n\nAll those actions for holding/result objects were implemented internally as plugins. You can right your own plugins, too.\n\nA plugin has to be defined as a class and then instantiated.\n\n**There are 5 APIs for developing a plugin for `cmdy`**\n\n- `cmdy._plugin_factory.register`: A decorator for the plugin class\n- `cmdy._plugin_factory.hold_then`: A decorator to decorate methods in the plugin class, which define actions after a holding object. Arguments:\n  - `alias`: The alias of this action (e.g. `r/redir` for `redirect`)\n  - `final`: Whether this is a final action, meaning no other actions should be followed\n  - `prop`: Whether this action can be called as a property\n  - `hold_right`: Should I put right following action on hold? This is useful when we have connectors which then can set the command running. (e.g `>` for redirect and `|` for pipe)\n- `cmdy._plugin_factory.run_then`: A decorator to decorate methods in the plugin class, which define actions after a sync result object. Arguments are similar as `cmdy._plugin_factory.hold_then` except that `prop` and `hold_right` are not avaialbe.\n- `async_run_then.add_method`: A decorator to decorate methods in the plugin class, which add methods to the `CmdyHolding`, `CmdyResult` or `CmdyAsyncResult` class. `cls` is the only argument that specifies which class we are hacking.\n- `async_run_then.add_property`: Property version of `cmdy_plugin_add_method`\n\n**Notes on name conflicts:**\n\nIf we need to add the methods to multiple classes in the plugin with the same name, you can define a different name with extra underscore suffix(es).\n\n**Notes on module baking:**\n\n- Always use the baked module to get those classes\n    ```pytho\n    import cmdy\n    cmdy2 = cmdy()\n\n    @cmdy2._plugin_factory.register\n    class MyPlugin:\n        ...\n    ```\n- Plugin enable and disable only take effect within the same module. For example:\n\n    ```python\n    import cmdy\n    sh = cmdy()\n    # only affects cmdy not sh\n    sh._plugins.fg.disable()\n    # to disable this plugin for sh as well:\n    sh._plugins.fg.disable()\n    ```\n\n```python\n# An example to define a plugin\nimport cmdy\n\n@cmdy._plugin_factory.register\nclass MyPlugin:\n    @cmdy._plugin_factory.add_method(cmdy.CmdyHolding)\n    def say_hello(self):\n        return \'Hello world!\'\n\n    @cmdy._plugin_factory.hold_then(\'hello\')\n    def helloworld(self):\n        print(self.say_hello())\n        # keep chaining\n        return self\n\nmyplugin = MyPlugin()\n\n# command will never run,\n# because we didn\'t do self.run() in helloworld(self)\nls().helloworld()  # prints Hello world!\n# property calls enabled by default\nls().helloworld  # prints Hello world!\n# we have alias\nls().hello  # prints Hello world!\n\n```\n\n[1]: https://img.shields.io/pypi/v/cmdy?style=flat-square\n[2]: https://pypi.org/project/cmdy/\n[3]: https://img.shields.io/github/tag/pwwang/cmdy?style=flat-square\n[4]: https://github.com/pwwang/cmdy\n[5]: https://img.shields.io/travis/pwwang/cmdy?style=flat-square\n[6]: https://travis-ci.org/pwwang/cmdy\n[7]: https://img.shields.io/codacy/grade/fa12f06d39404f98b94c19e83865fd4e?style=flat-square\n[8]: https://app.codacy.com/project/pwwang/cmdy/dashboard\n[9]: https://img.shields.io/codacy/coverage/fa12f06d39404f98b94c19e83865fd4e?style=flat-square\n[10]: https://img.shields.io/pypi/pyversions/cmdy?style=flat-square\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pwwang/cmdy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
